@@ -1,4 +1,4 @@
-use std::fmt;
+use std::{cmp::Ordering, fmt};
 
 use lazy_static::lazy_static;
 use regex::Regex;
@@ -103,7 +103,7 @@ impl fmt::Display for InvalidEnvironment {
 }
 
 /// Represents a parsed version.
-#[derive(Debug, Clone, PartialEq)]
+#[derive(Debug, Clone)]
 pub struct Version<'a> {
     raw: &'a str,
     major: &'a str,
@@ -111,6 +111,7 @@ pub struct Version<'a> {
     patch: &'a str,
     revision: &'a str,
     pre: &'a str,
+    before_code: &'a str,
     build_code: &'a str,
     components: u8,
 }
@@ -128,6 +129,7 @@ impl<'a> Serialize for Version<'a> {
         state.serialize_field("revision", &self.revision())?;
         state.serialize_field("pre", &self.pre())?;
         state.serialize_field("build_code", &self.build_code())?;
+        state.serialize_field("raw_short", &self.raw_short())?;
         state.serialize_field("components", &self.components())?;
         state.serialize_field("raw_quad", &self.raw_quad())?;
         state.end()
@@ -162,6 +164,11 @@ impl<'a> Version<'a> {
             return Err(InvalidVersion);
         }
 
+        let before_code = match caps.get(6) {
+            Some(cap) => &version[..cap.start() - 1],
+            None => version,
+        };
+
         Ok(Version {
             raw: version,
             major: caps.get(1).map(|x| x.as_str()).unwrap_or_default(),
@@ -178,6 +185,7 @@ impl<'a> Version<'a> {
                     pre
                 })
                 .unwrap_or(""),
+            before_code,
             build_code: caps.get(6).map(|x| x.as_str()).unwrap_or(""),
             components,
         })
@@ -264,6 +272,16 @@ impl<'a> Version<'a> {
         self.raw
     }
 
+    /// Returns the part of the version raw before the build code.
+    ///
+    /// This is useful as the system can mis-parse some versions and
+    /// instead of formatting out the version from the parts, this can
+    /// be used to format out the version part as it was input by the
+    /// user but still abbreviate the build code.
+    pub fn raw_short(&self) -> &str {
+        self.before_code
+    }
+
     /// Returns the version triple (major, minor, patch)
     pub fn triple(&self) -> (u64, u64, u64) {
         (self.major(), self.minor(), self.patch())
@@ -285,15 +303,57 @@ impl<'a> Version<'a> {
     }
 }
 
+impl<'a> Ord for Version<'a> {
+    fn cmp(&self, other: &Self) -> Ordering {
+        match self.quad().cmp(&other.quad()) {
+            Ordering::Equal => {
+                // handle pre-releases first
+                match (self.pre(), other.pre()) {
+                    (None, Some(_)) => return Ordering::Greater,
+                    (Some(_), None) => return Ordering::Less,
+                    (Some(self_pre), Some(other_pre)) => {
+                        match self_pre.cmp(&other_pre) {
+                            Ordering::Equal => {}
+                            other => return other,
+                        };
+                    }
+                    (None, None) => {}
+                }
+
+                // if we have build numbers, compare them
+                if let (Some(self_num), Some(other_num)) =
+                    (self.build_number(), other.build_number())
+                {
+                    return self_num.cmp(&other_num);
+                }
+
+                // lastly compare build code lexicographically
+                self.build_code().cmp(&other.build_code())
+            }
+            other => other,
+        }
+    }
+}
+
+impl<'a> PartialOrd for Version<'a> {
+    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
+        Some(self.cmp(other))
+    }
+}
+
+impl<'a> PartialEq for Version<'a> {
+    fn eq(&self, other: &Self) -> bool {
+        self.quad() == other.quad()
+            && self.pre() == other.pre()
+            && self.build_code() == other.build_code()
+    }
+}
+
+impl<'a> Eq for Version<'a> {}
+
 impl<'a> fmt::Display for Version<'a> {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        fmt::Display::fmt(&VersionDescription(self), f)?;
-        if let Some(pre) = self.pre() {
-            write!(f, "-{}", pre)?;
-        }
-        if let Some(build_code) = self.build_code() {
-            write!(f, "+{}", build_code)?;
-        }
+        write!(f, "{}", self.raw)?;
         Ok(())
     }
 }
@@ -437,30 +497,6 @@ impl<'a> Release<'a> {
     }
 }
 
-#[derive(Debug)]
-struct VersionDescription<'a>(&'a Version<'a>);
-
-impl<'a> fmt::Display for VersionDescription<'a> {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        match self.0.raw_quad() {
-            (major, Some(minor), Some(patch), Some(revision)) => {
-                write!(f, "{}.{}.{}.{}", major, minor, patch, revision)?;
-            }
-            (major, Some(minor), Some(patch), None) => {
-                write!(f, "{}.{}.{}", major, minor, patch,)?;
-            }
-            (major, Some(minor), None, None) => {
-                write!(f, "{}.{}", major, minor)?;
-            }
-            (major, None, None, None) => {
-                write!(f, "{}", major)?;
-            }
-            _ => unreachable!(),
-        }
-        Ok(())
-    }
-}
-
 /// Helper object to format a release into a description.
 #[derive(Debug)]
 pub struct ReleaseDescription<'a>(&'a Release<'a>);
@@ -473,10 +509,7 @@ impl<'a> fmt::Display for ReleaseDescription<'a> {
             .map(|hash| hash.get(..12).unwrap_or(hash));
 
         if let Some(ver) = self.0.version() {
-            fmt::Display::fmt(&VersionDescription(ver), f)?;
-            if let Some(pre) = ver.pre() {
-                write!(f, "-{}", pre)?;
-            }
+            write!(f, "{}", ver.raw_short())?;
             if let Some(short_hash) = short_hash {
                 write!(f, " ({})", short_hash)?;
             } else if let Some(build_code) = ver.build_code() {
@@ -552,3 +585,38 @@ fn test_environment_validation() {
     );
     assert_eq!(validate_environment("good"), Ok(()));
 }
+
+#[test]
+fn test_version_ordering() {
+    macro_rules! ver {
+        ($v:expr) => {
+            Version::parse($v).unwrap()
+        };
+    }
+
+    assert_eq!(ver!("1.0.0"), ver!("1.0.0"));
+    assert_ne!(ver!("1.1.0"), ver!("1.0.0"));
+    assert_eq!(ver!("1.0"), ver!("1.0.0"));
+    assert_eq!(ver!("1.0dev"), ver!("1.0.0-dev"));
+    assert_eq!(ver!("1.0dev"), ver!("1.0.0.0-dev"));
+    assert_ne!(ver!("1.0dev"), ver!("1.0.0.0-dev1"));
+
+    assert!(ver!("1.0dev") >= ver!("1.0.0.0-dev"));
+    assert!(ver!("1.0dev") < ver!("1.0.0.0"));
+
+    assert!(ver!("1.0.0") > ver!("1.0.0-rc1"));
+    assert!(ver!("1.0.0") >= ver!("1.0.0-rc1"));
+    assert!(ver!("1.0.0-rc1") > ver!("0.9"));
+    assert!(ver!("1.0.0+10") < ver!("1.0.0+20"));
+    assert!(ver!("1.0.0+a") < ver!("1.0.0+b"));
+
+    assert!(ver!("1.0") < ver!("2.0"));
+    assert!(ver!("1.1.0") < ver!("10.0"));
+    assert!(ver!("1.1.0") < ver!("1.1.1"));
+    assert!(ver!("1.1.0.1") < ver!("1.1.0.2"));
+    assert!(ver!("1.1.0.1") > ver!("1.1.0.0"));
+    assert!(ver!("1.1.0.1") > ver!("1.0.0.0"));
+    assert!(ver!("1.1.0.1") > ver!("1.0.42.0"));
+
+    assert!(ver!("1.0+abcd") < ver!("1.0+abcde"));
+}
